const {isValue} = require('@icode-run/utils/cjs/utils');
const {getOutputValue} = require('@icode-run/utils/cjs/dataType');
const request = require("request");
const fs = require("fs");
const path = require("path");

/**
 * 把组件原配置中的staticData转化成组件props.data数据
 * @param {*} dataDescription 静态数据（表格数据）
 * @returns 返回转化后的数据，用于组件显示
 */
const _getStaticData = (dataDescription) => {
    const staticData = [];
    dataDescription.forEach((rowData, rowIndex) => {
      if(rowIndex > 0) {
        const rowItem = {};
        (rowData || []).map((colData, colIndex) => {
          rowItem[colIndex + ''] = colData;
        })
        staticData.push(rowItem);
      }
    })
  
    return staticData;
  }
exports.getStaticData = _getStaticData;

const _getDefaultValue = (keyName, defaultValueConfig) => {
  if (!(defaultValueConfig && defaultValueConfig.ruleType && isValue(defaultValueConfig.ruleValue))) {
    return;
  }
  let value = "", type = "";
  if (defaultValueConfig.ruleType === "static") {
    if (
      Number(defaultValueConfig.ruleValue) ==
      defaultValueConfig.ruleValue
    ) {
      type = "number";
    } else {
      type = "string";
    }
    value = defaultValueConfig.ruleValue;
  } else {
    value = `getDefaultValue('${defaultValueConfig.ruleType}', '${defaultValueConfig.ruleValue}')`;
    type = "fn";
  }
  return {
    keyName: keyName,
    value: value,
    type: type,
  }
}
exports.getDefaultValue = _getDefaultValue;

/**
 * 获取初始化组件的默认值
 * ——静态数据(state.DefaultValue = currRelation.defaultValueConfig)
 * ——动态数据, 未给param绑定参数项(state.DefaultValue = currRelation.defaultValueConfig)
 * ——动态数据，已给param绑定参数项(state.Params = currRelation.defaultValueConfig
 * @param {*} propsOption 
 * @param {*} currRelation 
 */
exports.getParamDefaultValueState = ({
  paramStates, 
  propsOptions, 
  currRelation,
  relations,
  dataModels,
}) => {
  paramStates = paramStates || {};
  const staticDefaultValue = currRelation.defaultValueConfig || {};
  
  if(!Object.keys(staticDefaultValue).length) {
    return;
  }
  (propsOptions || []).forEach(options => {
    if(options.type === 'param' && staticDefaultValue[options.name]) {
      if(options.value && options.value.length) {
        options.value.map(valueItem => {
          const from = dataModels[valueItem.serviceName] && dataModels[valueItem.serviceName].type;
          const paramsName = `${exports.getApiStateName(valueItem.nodeId, {relations, dataModels, from})}Params`;
          paramStates[paramsName] = paramStates[paramsName] || {};
          let paramValue = _getDefaultValue(valueItem.keyName, staticDefaultValue[options.name]);
          const newParamValue = getOutputValue({
            outputType: valueItem.structureType,
            value: paramValue.value,
            ...(valueItem.transformType && isValue(valueItem.transformType.index) ? {
              targetIndex: valueItem.transformType.index
            }:{}),
            ...(valueItem.transformType && valueItem.transformType.join ? {
              valueSplit: valueItem.transformType.join
            }:{}),
          })
          paramStates[paramsName][valueItem.keyName] = {
            keyName: valueItem.keyName,
            value: newParamValue,
            type: valueItem.structureType,
          };
        })
        
      }
      const defaultName = `${exports.getApiStateName(currRelation.nodeId, {relations, dataModels})}DefaultValue`;
      paramStates[defaultName] = paramStates[defaultName] || {};
      paramStates[defaultName][options.name] = {
        ..._getDefaultValue(options.name, staticDefaultValue[options.name]),
        type: options.outputType,
      };
      
    }
  })

}

/**
 * 
 */
exports.getNewEvents = ({
  events = [], 
  dataModels,
  relations,
  paramDefaultValueState = {},
}) => {
  const currentEvents = {};
  let stateNodeId;
  events.map((item) => {
    if (item.eventName && item.eventType && item.eventContent) {
      currentEvents[item.eventName] = currentEvents[item.eventName] || [];
      if(item.stateNodeId) {
        stateNodeId = exports.getApiStateName(item.stateNodeId, {dataModels, relations});
      }
      const relationService = [];
      Object.keys(dataModels).forEach(dataModelName => {
        const dataModel = dataModels[dataModelName] || {};
        const apiParams = dataModel.apiParams || {};
        Object.keys(apiParams).forEach(keyName => {
          const defaultValueConfig = (apiParams[keyName] && apiParams[keyName].defaultValueConfig) || {};
          if(defaultValueConfig.ruleType === 'event' && defaultValueConfig.ruleValue && defaultValueConfig.ruleValue[0] === item.stateNodeId) {
            const serviceName = exports.getApiServiceName(dataModel);
            if(serviceName) {
              relationService.push({
                serviceName,
                keyName: defaultValueConfig.ruleValue[1]
              });
            }
          }
        })
      })
      
      let eventContent = item.eventContent;
      if(['openModal', 'closeModal'].indexOf(item.eventType) >= 0) {
        eventContent = `${exports.getApiStateName(item.eventContent, {relations, dataModels})}Visible`;
      }
      
      if('refresh' === item.eventType) {
        const nodeId = item.eventContent;
        const relation = relations[nodeId];
        const dataModel = dataModels[relation.dataSourceName] || {};
        if(relation.dataSourceType === 'api' && relation.dataSourceName && dataModels[relation.dataSourceName]) {
          const paramStateName = `${exports.getApiStateName(nodeId, {
            relations,
            dataModels,
            from: dataModel.type,
          })}Params`;
          const currParamsDefaultValue = paramDefaultValueState[paramStateName] || {};
          const params = {};
          if(item.resetParams && item.resetParams.length) {
            item.resetParams.forEach(keyName => {
              params[keyName] = currParamsDefaultValue[keyName] || currParamsDefaultValue[keyName] === 0 ? currParamsDefaultValue[keyName] : '';
            })
          }
          relationService.push({
            serviceName: exports.getApiServiceName(dataModel),
            params: params
          })
        }
      }

      currentEvents[item.eventName].push({
        eventType: item.eventType,
        ...(eventContent ? {eventContent} : {}),
        ...(relationService.length ? {
          relationService,
        } : {}),
        ...(item.stateNodeId ? {
          stateNodeId: `${exports.getApiStateName(item.stateNodeId, {relations, dataModels})}EventData`
        } : {}),
        ...(item.messageContent ? {
          messageContent: item.messageContent
        } : {}),
        ...(item.messageDuration ? {
          messageDuration: item.messageDuration
        } : {}),
      });
    }
  });
  return {
    events: currentEvents,
    stateNodeId
  };
}

/**
 * 给dataModels更新dataKeyName的值，优化了构建后的函数名称
 * 返回无重复的dataModel，以便优化构建service.ejs
 */
let dataModelsNames = {};
exports.setNewDataModelName = (dataModels) => {
  let key;
  const newDataModels = [];
  for(key in dataModels) {
    const name = dataModels[key].name;
    if(!dataModelsNames[name]) {
      newDataModels.push(dataModels[key]);
    }
    dataModels[key].dataKeyName = name;
    dataModels[key].dataModelName = key;
    dataModelsNames[name] = (dataModelsNames[name] || 0) + 1;
  }
  return newDataModels;
}

/**
 * 把旧的nodeId换成新的nodeId
 */
let nodeIds = {}, newNodeIdCount = {}, oldNodeIds = {};
exports.setNewNodeId = (item) => {
  const nodeId = item.nodeId;
  const componentName = item.componentName;
  newNodeIdCount[componentName] = (newNodeIdCount[componentName] || 0) + 1;
  nodeIds[nodeId] = `${componentName}${newNodeIdCount[componentName]}`;
  oldNodeIds[nodeIds[nodeId]] = nodeId;
  oldNodeIds[nodeId] = nodeId;
  return nodeIds[nodeId];
};

/**
 *  获取当前relation, 根据新/旧的nodeId，找到relation中旧的nodeId相关的数据;
 */
exports.getCurrentRelations = (nodeId, relations = {}) => {
  // if(relations[oldNodeIds]) {
  //   relations[oldNodeIds].nodeId = nodeId;
  // }
  return relations[oldNodeIds[nodeId] || nodeId] || {};
}

/**
 *  获取Api请求函数名称, 用旧的nodeId换新的nodeId
 */
exports.getApiServiceName = (dataModel = {}) => {
  if(!dataModel.nodeId) {
    return '';
  }
  const newNodeId = nodeIds[dataModel.nodeId] || dataModel.nodeId;
  return `${dataModel.type}${newNodeId}`;
};

/**
 *  获取Api相关数据的state名称, 用旧的nodeId换新的nodeId
 * from: 'update'-dataModel==='update'时；
 */
exports.getApiStateName = (nodeId, {
  dataModels = {}, 
  relations = {},
  from = '',
}) => {
  const newNodeId = nodeIds[nodeId] || nodeId || 'nodeId';
  return from === 'update' ? `update${newNodeId}` : `${newNodeId.charAt(0).toLowerCase() + newNodeId.slice(1)}`;
};

/**
 *  获取Api相关数据的state名称, 用旧的nodeId换新的nodeId
 * from: 'update'-dataModel==='update'时；
 */
 exports.getNewNodeId = (nodeId) => {
  const newNodeId = nodeIds[nodeId] || nodeId;
  return `${newNodeId.charAt(0).toLowerCase() + newNodeId.slice(1)}`;
};

/**
 *  接口请求成功后，更新相关联的“其它接口参数默认值”的当前值
 *  return 需要更新的参数字符串
 */
exports.getNeedUpdateParams = (currDataModelName, {
  dataModels = {},
}) => {
  let needUpdateParams = '';
  Object.keys(dataModels).forEach(dataModelsName => {
    const dataModel = dataModels[dataModelsName] || {};
    const apiParams = dataModel.apiParams || {};
    let paramsString = '';
    Object.keys(apiParams).forEach(keyName => {
      const defaultValueConfig = (apiParams[keyName] && apiParams[keyName].defaultValueConfig) || {};
      if(
        defaultValueConfig.ruleType === 'api' && 
        defaultValueConfig.ruleValue &&  
        defaultValueConfig.ruleValue[0] === currDataModelName && 
        defaultValueConfig.ruleValue[1] 
      ) {
        paramsString += `${keyName}: data['${defaultValueConfig.ruleValue[1]}'],\n`;
      }
    })
    if(paramsString) {
      needUpdateParams += `${exports.getApiStateName(dataModel.nodeId, {from: dataModel.type})}Params: {
        ${paramsString}
      },\n`
    }
  })
  return needUpdateParams;
}

exports.getCss = (cssContent) => {
  const splitNames = cssContent.match(/[a-zA-Z0-Z&-_>: ]*\s*\{/g);
  const indexs = splitNames.map((name) => {
      return cssContent.indexOf(name)
  });
  let currCss = [], otherCss = [];
  indexs.forEach((indexName, index) => {
    const isCurr = splitNames[index].match(/^\&/);
    if(isCurr && isCurr.length && indexs[index + 1]) {
        currCss.push(cssContent.substring(indexName, indexs[index + 1] - 1));
    }else {
        otherCss.push(cssContent.substring(indexName, indexs[index + 1] ? indexs[index + 1] - 1 : cssContent.length));
    }
  })
  return {
    '&': currCss.join('\n'),
    'global': otherCss.join('\n'),
  }
}

exports.getStaticDataName = (nodeId) => {
  return `staticData${nodeId}`;
}

exports.getRefreshDataHtml = (refreshData) => {
  let timeing = 1;
  const apis = {};
  const statics = {};
  refreshData.forEach(item => {
    if(item.refreshData && item.refreshData.timing) {
      timeing = item.refreshData.timing > timeing ? item.refreshData.timing : timeing;
    }
    if(item.type === 'api' && item.serviceName) {
      apis[item.serviceName] = `this.${item.serviceName}()`;
    }
    if(item.type === 'static' && item.stateName && item.refreshData && item.refreshData.config) {
      statics[item.stateName] = `refreshStaticData(this.state.${item.stateName},${JSON.stringify(item.refreshData.config)}),`;
    }
  })
  const staticsParams = Object.keys(statics).map(stateName => {
    return `${stateName}: ${statics[stateName]}`
  });
  if(Object.values(apis).length || staticsParams.length) {
    return `
  refreshTimeValue = null;
  refreshData = () => {
    if(this.refreshTimeValue) {
      clearTimeout(this.refreshTimeValue);
      this.refreshTimeValue = null;
    }
    this.refreshTimeValue = setTimeout(() => {
      clearTimeout(this.refreshTimeValue);
      this.refreshTimeValue = null;
      ${Object.values(apis).join('\n')}
      ${staticsParams.length ? ['this.setState({'].concat(staticsParams, ['});']).join('\n') : ''}

      this.refreshData();
    }, ${timeing * 1000});
  }
  `;
  }
  else {
    return '';
  }
}

exports.getRequestStructuresType = (dataModel, keyName) => {
  const requestStructures = dataModel.requestStructures || [];
  let type;
  requestStructures.forEach(item => {
    if(item.keyName === keyName) {
      type = item.type;
    }
  })
  return type;
}


const _saveImgToLocation = ({
  url,
  imagesDirPath
}) => {
  if(url.match(/^http(|s)\:\/\//) && imagesDirPath) {
    const filenames = url.split('/');
    const filename = `online_${filenames[filenames.length - 1]}`;
    if (!fs.existsSync(imagesDirPath)) {
      fs.mkdirSync(imagesDirPath);
    }
    request.get({uri:url, encoding:'binary'},function(err,res){
      if(!err) {
        fs.writeFile(path.join(imagesDirPath, filename),res.body,"binary",function(err,res){
          
        })
      }
    })
    return `/assets/${filename}`;
  }
  return url;
}
exports.saveImgToLocation = _saveImgToLocation;

exports.saveGlobalState = ({
  globalState = {},
  globalDirPath
}) => {
  if(globalState.isAppTheme && globalDirPath) {
    const filePath = path.join(globalDirPath, 'globalTheme.json');
    if (!fs.existsSync(globalDirPath)) {
      fs.mkdirSync(globalDirPath);
    }
    fs.writeFileSync(filePath,JSON.stringify(globalState),'utf-8');
    return {
      dependencies: "import globalTheme from '@/layouts/globalTheme.json';",
      content: 'globalTheme'
    }
  }

  return {
    content: JSON.stringify(globalState)
  };
}

exports.replaceImgUrlInCss =  (content, imagesDirPath) => {
  if(!imagesDirPath) {
    return content;
  }
  let newContent = content.replace(/url\(('|")*\/staticImages\/(.(?!url))*\)/g, (value) => {
    let url = value.replace(/^url\(('|")*/, '');
    url = url.replace(/('|")*\)$/, '');
    url = _saveImgToLocation({
      url: `https://www.icoderun.com${url}`,
      imagesDirPath,
    })
    return `url(${url})`;
  });
  return newContent;
}

exports.replaceImgUrlInProps = (url = '', imagesDirPath) => {
  if(!imagesDirPath || !url) {
    return url;
  }
  let newUrl;
  let isSaveImg = false;
  if(url.match(/^\/staticImages\//)) {
    isSaveImg = true;
    newUrl = `https://www.icoderun.com${url}`;
  }
  else if(url.match(/^http:\/\/127\.0\.0\.1:\d+\/assets\//)) {
    newUrl = url ? url.replace(/http:\/\/127\.0\.0\.1:\d+\/assets\//, '/assets/') : '';
  }
  else if(url.match(/^http(|s)*\:\/\//)) {
    isSaveImg = true;
    newUrl = url;
  }
  if(isSaveImg) {
    newUrl = _saveImgToLocation({
      url: newUrl,
      imagesDirPath: imagesDirPath,
    })
  }
  
  return newUrl;
}

exports.buildComponentDependenciesHtml = ({
  dependencies = {},
  versions = {},
  remoteEntryPaths= {},
  isLoadingNpm = false,
  packagePath,
}) => {
  let html = '';
  const packageDependencies = {};
  let remoteApps = '';
  Object.keys(dependencies).forEach(componentName => {
    const item = dependencies[componentName] || {}
    if(isLoadingNpm) {
      html += `import ${item.componentName} from '${item.packageName}';\n`;
      packageDependencies[item.packageName] = `${versions[item.packageName] || '*'}`;
    }
    else if(remoteEntryPaths[item.remoteAppName]){
      remoteApps += `${item.remoteAppName}: '${remoteEntryPaths[item.remoteAppName]}',\n`;
      html += `const ${item.componentName} = importRemote('${item.remoteAppName}/${item.componentName}');\n`;
    }
  })

  if(isLoadingNpm && packagePath) {
    const filePath = path.join(packagePath, 'package.json');
    const content = fs.readFileSync(filePath, 'utf-8');
    const data = JSON.parse(content);
    data.dependencies = {
      ...data.dependencies,
      ...packageDependencies,
    };
    fs.writeFileSync(filePath, JSON.stringify(data), 'utf-8');
  }
  else if(remoteApps){
    let remoteEntry = `
      import importRemote, { setRemoteConfig } from 'import-remote-module';
      setRemoteConfig({
        ${remoteApps}
      });
    `;
    
    html = `${remoteEntry}
    ${html}`
  }
  else {
    html = '';
  }
  return html;
}

exports.init = () => {
  nodeIds = {};
  newNodeIdCount = {};
  oldNodeIds = {};
  dataModelsNames = {};
}